import numpy as np
import sys
import matplotlib.pyplot as plt

# Set this to True to use random matrices instead of extracting from memory
use_random = False

MEM_ROW = 20
MEM_COL = 100

TOP_K = 4  # Top-K attention entries to retain

# Matrix T config (from mem1)
T_start_row = 0
T_start_col = 0

# Matrix W config (from mem2)
W_start_row = 0
W_start_col = 0

# ------------------ Helper Functions ------------------

def read_flattened_hex_file(filepath, rows, cols):
    data = []
    with open(filepath, 'r') as f:
        for line in f:
            line = line.strip()
            if line == '':
                continue
            val = float(int(line, 16))  # hex -> int -> float
            data.append(val)

    if len(data) != rows * cols:
        raise ValueError(f"Data size mismatch: expected {rows*cols}, got {len(data)}")

    return np.array(data).reshape(rows, cols)

def extract_matrix(memory, start_row, start_col, num_rows, num_cols):
    return memory[start_row:start_row + num_rows, start_col:start_col + num_cols]

def random_project(matrix, s):
    L, D = matrix.shape
    d = D // 4
    R = np.random.choice([1, -1, 0], size=(D, d), p=[1 / (2 * s), 1 / (2 * s), 1 - 1 / s])
    return np.dot(matrix, R)

def get_model_params(choice):
    model_configs = {
        1: {"name": "BERT-Base",  "L": 512, "D": 768, "d": 128, "s": 64, "H": 12},
        2: {"name": "BERT-Large", "L": 512, "D": 1024, "d": 256, "s": 128, "H": 16},
        3: {"name": "ViT-B",      "L": 197, "D": 768, "d": 128, "s": 64, "H": 12},
        4: {"name": "ViT-L",      "L": 197, "D": 1024, "d": 256, "s": 128, "H": 16},
        5: {"name": "Trial1",     "L": 16, "D": 30, "d": 8, "s": 2, "H": 2},
        6: {"name": "Trial2",     "L": 16, "D": 12, "d": 3, "s": 2, "H": 2},
        7: {"name": "Trial3",     "L": 20, "D": 16, "d": 4, "s": 2, "H": 2},
    }
    return model_configs.get(choice, None)

def apply_topk_bitmask(A, top_k):
    mask = np.zeros_like(A, dtype=bool)
    for i in range(A.shape[0]):
        topk_indices = np.argsort(A[i])[::-1][:top_k]
        mask[i, topk_indices] = True
    return mask

def softmax(x):
    x_max = np.max(x, axis=-1, keepdims=True)
    e_x = np.exp(x - x_max)
    return e_x / np.sum(e_x, axis=-1, keepdims=True)

# ------------------ Main Code ------------------

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python3 RPAS_algorithm.py <model_choice>")
        sys.exit(1)

    try:
        choice = int(sys.argv[1].strip())
        params = get_model_params(choice)

        if params:
            mem1_path = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem3.hex"#mem1.hex"
            mem2_path = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem3.hex"#mem2.hex"

            memory1 = read_flattened_hex_file(mem1_path, MEM_ROW, MEM_COL)
            memory2 = read_flattened_hex_file(mem2_path, MEM_ROW, MEM_COL)

            if use_random:
                T = np.random.rand(params["L"], params["D"])  # Random matrix of shape (L, D)
                W = np.random.rand(params["D"], params["D"])  # Random matrix of shape (D, D)
            else:
                T = extract_matrix(memory1, T_start_row, T_start_col, params["L"], params["D"])
                W = extract_matrix(memory2, W_start_row, W_start_col, params["D"], params["D"])

            print("✅ Token matrix T:")
            print(T)
            print("Shape:", T.shape)

            print("\n✅ Weight matrix W:")
            print(W)
            print("Shape:", W.shape)

            D_h = params["D"] // params["H"]

            Q_list = []
            K_list = []
            A_list = []
            MaskA = {}
            MaskQ = {}
            MaskKV = {}
            Softmax_list = []
            Masked_softmax_list = []
            Diff_list = []
            Accuracy_list = []

            for h in range(params["H"]):
                start = h * D_h
                end = (h + 1) * D_h

                # Full Q and K (not projected!)
                WQ_h = W[:, start:end]
                WK_h = W[:, start:end]

                Q = np.dot(T, WQ_h)
                K = np.dot(T, WK_h)

                print(f"\n🎯 Head {h} Q shape: {Q.shape}, K shape: {K.shape}")

                A = np.dot(Q, K.T)

                print(f"🧮 Head {h} Raw Attention Scores (A):")
                print(A)

                mask = apply_topk_bitmask(A, TOP_K)
                MaskA[h] = mask
                MaskQ[h] = np.any(mask, axis=1)
                MaskKV[h] = np.any(mask, axis=0)

                A_softmax = softmax(A)

                A_masked = np.where(mask, A, -1e9)
                A_masked_softmax = softmax(A_masked)

                diff = np.abs(A_softmax - A_masked_softmax)
                relative_error = np.sum(diff) / np.sum(A_softmax)
                accuracy = 100 * (1 - relative_error)

                print(f"🎯 Head {h} Accuracy: {accuracy:.2f}%")

                Q_list.append(Q)
                K_list.append(K)
                A_list.append(A)
                Softmax_list.append(A_softmax)
                Masked_softmax_list.append(A_masked_softmax)
                Diff_list.append(diff)
                Accuracy_list.append(accuracy)

            print(f"\n📘 Model: {params['name']}")
            print(f"L: {params['L']}, D: {params['D']}, d: {params['d']}, H: {params['H']}, TopK: {TOP_K}")
            

            avg_accuracy = np.mean(Accuracy_list)
            print(f"\n📊 Average Attention Accuracy Across All Heads: {avg_accuracy:.2f}%")

            # # Computational savings
            # L = params["L"]
            # D_h = params["D"] // params["H"]
            # flops_full = L * L * D_h
            # flops_topk = L * TOP_K * D_h
            # total_full = params["H"] * flops_full
            # total_topk = params["H"] * flops_topk
            # overall_saving = 100 * (1 - total_topk / total_full)

            # print(f"\n⚡ Estimated Computational Savings (TopK vs Full Attention): {overall_saving:.2f}%")



            # Show results for one head
            head_to_display = 1
            if head_to_display in MaskA:
                # print(f"\n🧠 Head {head_to_display} - TopK Attention Mask (MaskA):")
                # print(MaskA[head_to_display].astype(int))

                # print(f"\nMaskQ[{head_to_display}]:")
                # print(MaskQ[head_to_display].astype(int))

                # print(f"\nMaskKV[{head_to_display}]:")
                # print(MaskKV[head_to_display].astype(int))

                # print(f"\n🔎 Softmax Attention Matrix (Original) - Head {head_to_display}:")
                # print(Softmax_list[head_to_display])

                # print(f"\n🔒 Masked Softmax Attention Matrix - Head {head_to_display}:")
                # print(Masked_softmax_list[head_to_display])

                # print(f"\n🔍 Difference Matrix - Head {head_to_display}:")
                # print(Diff_list[head_to_display])

                # Plot heatmap
                plt.figure(figsize=(10, 6))
                plt.imshow(Diff_list[head_to_display], cmap='viridis', interpolation='nearest')
                plt.colorbar(label="Difference Magnitude")
                plt.title(f"Head {head_to_display} - Softmax Difference (Original vs Masked)")
                plt.xlabel("Key Index")
                plt.ylabel("Query Index")
                plt.tight_layout()
                plt.show()

            else:
                print(f"\nWarning: Head {head_to_display} not found in computed masks.")

        else:
            print("\nInvalid model choice! Please use a number from 1 to 7.")

    except ValueError as e:
        print("\nError:", e)
