import numpy as np
import sys
import matplotlib.pyplot as plt

MEM_ROW = 20
MEM_COL = 100
TOP_K = 4
H_MASK_STEP_SIZE = 4  # New constant for mask location step per head

# Matrix T config (from mem1)
T_start_row = 0
T_start_col = 0

# Matrix W config (from mem2)
W_start_row = 0
W_start_col = 0

# ------------------ Helper Functions ------------------

def read_flattened_hex_file(filepath, rows, cols):
    data = []
    with open(filepath, 'r') as f:
        for line in f:
            line = line.strip()
            if line == '':
                continue
            val = float(int(line, 16))
            data.append(val)

    if len(data) != rows * cols:
        raise ValueError(f"Data size mismatch: expected {rows*cols}, got {len(data)}")

    return np.array(data).reshape(rows, cols)

def extract_matrix(memory, start_row, start_col, num_rows, num_cols):
    return memory[start_row:start_row + num_rows, start_col:start_col + num_cols]

def get_model_params(choice):
    model_configs = {
        1: {"name": "BERT-Base",  "L": 512, "D": 768, "d": 128, "s": 64, "H": 12},
        2: {"name": "BERT-Large", "L": 512, "D": 1024, "d": 256, "s": 128, "H": 16},
        3: {"name": "ViT-B",      "L": 197, "D": 768, "d": 128, "s": 64, "H": 12},
        4: {"name": "ViT-L",      "L": 197, "D": 1024, "d": 256, "s": 128, "H": 16},
        5: {"name": "Trial1",     "L": 16, "D": 30, "d": 8, "s": 2, "H": 2},
        6: {"name": "Trial2",     "L": 16, "D": 12, "d": 3, "s": 2, "H": 2},
        7: {"name": "Trial3",     "L": 20, "D": 16, "d": 4, "s": 2, "H": 2},
    }
    return model_configs.get(choice, None)

def apply_topk_bitmask(A, top_k):
    mask = np.zeros_like(A, dtype=bool)
    for i in range(A.shape[0]):
        topk_indices = np.argsort(A[i])[::-1][:top_k]
        mask[i, topk_indices] = True
    return mask

def softmax(x):
    x_max = np.max(x, axis=-1, keepdims=True)
    e_x = np.exp(x - x_max)
    return e_x / np.sum(e_x, axis=-1, keepdims=True)

# ------------------ Read Packed Mask from Memory ------------------

def read_memory(file_path):
    with open(file_path, 'r') as f:
        content = f.read().split()
    return np.array([int(val, 16) for val in content], dtype=np.uint16).reshape((MEM_ROW, MEM_COL))

def read_mask_A_from_memory_per_row(mem, head_index, base_col=92, rows=20):
    """Reads 2 columns of packed mask per head based on head_index and base_col."""
    start_col = base_col + head_index * H_MASK_STEP_SIZE
    packed_values = []
    for r in range(rows):
        packed_values.append(mem[r, start_col])
        packed_values.append(mem[r, start_col + 1])
    return np.array(packed_values, dtype=np.uint16)

def unpack_mask_rows(packed, num_rows=20, num_cols=20):
    mask = []
    for i in range(0, len(packed), 2):
        lsb = f"{packed[i]:016b}"
        msb = f"{packed[i+1]:04b}"
        bitstring = msb + lsb
        row_mask = [int(b) for b in bitstring[:num_cols]]
        row_mask.reverse()
        mask.append(row_mask)
    return np.array(mask, dtype=np.uint8)

# ------------------ Main ------------------

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Usage: python3 RPAS_algorithm.py <model_choice>")
        sys.exit(1)

    try:
        choice = int(sys.argv[1].strip())
        params = get_model_params(choice)

        if params:
            mem_path_T = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem1.hex"
            mem_path_W = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem2.hex"
            mem_path_M = "/mnt/d/Project_verilog_item/project_files/hdl_verilog/phase5/file_dump/mem3.hex"

            # Read memories
            memory1 = read_flattened_hex_file(mem_path_T, MEM_ROW, MEM_COL)
            memory2 = read_flattened_hex_file(mem_path_W, MEM_ROW, MEM_COL)
            memory_raw = read_memory(mem_path_M)

            # Extract matrices
            T = extract_matrix(memory1, T_start_row, T_start_col, params["L"], params["D"])
            W = extract_matrix(memory2, W_start_row, W_start_col, params["D"], params["D"])

            # Set print options for readability
            np.set_printoptions(precision=3, suppress=True, threshold=1000)

            # Print matrices and masks
            print("✅ Token matrix T (shape {}):".format(T.shape))
            print(T)

            print("\n✅ Weight matrix W (shape {}):".format(W.shape))
            print(W)

            D_h = params["D"] // params["H"]
            Accuracy_list = []
            Memory_Mask_Accuracy_list = []

            for h in range(params["H"]):
                print(f"\n🎯 Processing Head {h}:")

                start = h * D_h
                end = (h + 1) * D_h

                WQ_h = W[:, start:end]
                WK_h = W[:, start:end]

                Q = np.dot(T, WQ_h)
                K = np.dot(T, WK_h)

                print(f"\n- Q matrix for head {h} (shape {Q.shape}):")
                print(Q)

                print(f"\n- K matrix for head {h} (shape {K.shape}):")
                print(K)

                A = np.dot(Q, K.T)
                print(f"\n- Attention scores matrix A (Q x K^T) for head {h} (shape {A.shape}):")
                print(A)

                A_softmax = softmax(A)
                print(f"\n- Softmax of A for head {h}:")
                print(A_softmax)

                # --- Top-K based masking (baseline)
                mask_topk = apply_topk_bitmask(A, TOP_K)
                print(f"\n- Top-{TOP_K} mask for head {h} (boolean matrix):")
                print(mask_topk.astype(int))

                A_masked = np.where(mask_topk, A, -1e9)
                print(f"\n- Masked A with Top-K mask applied (values masked with -1e9):")
                print(A_masked)

                A_masked_softmax = softmax(A_masked)
                print(f"\n- Softmax after applying Top-K mask:")
                print(A_masked_softmax)

                diff_topk = np.abs(A_softmax - A_masked_softmax)
                print(f"\n- Absolute difference between original softmax and Top-K masked softmax:")
                print(diff_topk)

                relative_error_topk = np.sum(diff_topk) / np.sum(A_softmax)
                accuracy_topk = 100 * (1 - relative_error_topk)
                Accuracy_list.append(accuracy_topk)
                print(f"\n🔹 TopK Mask Accuracy Calculation Steps:")
                print(f"  Sum of absolute differences: {np.sum(diff_topk):.6f}")
                print(f"  Sum of original softmax elements: {np.sum(A_softmax):.6f}")
                print(f"  Relative error: {relative_error_topk:.6f}")
                print(f"  Accuracy (%): {accuracy_topk:.2f}%")

                # --- Read memory-based mask (dynamic column based on head index)
                mem_mask_packed = read_mask_A_from_memory_per_row(memory_raw, h)
                print(f"\n📦 Packed Mask from memory (raw uint16 values) for Head {h}:")
                print(mem_mask_packed)

                mem_mask = unpack_mask_rows(mem_mask_packed, num_rows=20, num_cols=20).astype(bool)
                print(f"\n🧩 Expanded Mask matrix A from memory for Head {h} (shape {mem_mask.shape}):")
                print(mem_mask.astype(int))

                mem_mask_cropped = mem_mask[:params["L"], :params["L"]]
                print(f"\n- Memory-based mask (cropped to A shape {mem_mask_cropped.shape}):")
                print(mem_mask_cropped.astype(int))

                if mem_mask_cropped.shape != A.shape:
                    print(f"⚠️ Warning: Memory mask shape {mem_mask_cropped.shape} does not match attention matrix shape {A.shape}. Adjusting by padding/truncating.")
                    rows = min(mem_mask_cropped.shape[0], A.shape[0])
                    cols = min(mem_mask_cropped.shape[1], A.shape[1])
                    mem_mask_cropped = mem_mask_cropped[:rows, :cols]
                    A = A[:rows, :cols]
                    A_softmax = A_softmax[:rows, :cols]

                A_masked_mem = np.where(mem_mask_cropped, A, -1e9)
                print(f"\n- Masked A with Memory-based mask applied (values masked with -1e9):")
                print(A_masked_mem)

                A_masked_mem_softmax = softmax(A_masked_mem)
                print(f"\n- Softmax after applying Memory-based mask:")
                print(A_masked_mem_softmax)

                diff_mem = np.abs(A_softmax - A_masked_mem_softmax)
                print(f"\n- Absolute difference between original softmax and Memory Masked softmax:")
                print(diff_mem)

                relative_error_mem = np.sum(diff_mem) / np.sum(A_softmax)
                accuracy_mem = 100 * (1 - relative_error_mem)
                Memory_Mask_Accuracy_list.append(accuracy_mem)
                print(f"\n🔹 Memory Mask Accuracy Calculation Steps:")
                print(f"  Sum of absolute differences: {np.sum(diff_mem):.6f}")
                print(f"  Sum of original softmax elements: {np.sum(A_softmax):.6f}")
                print(f"  Relative error: {relative_error_mem:.6f}")
                print(f"  Accuracy (%): {accuracy_mem:.2f}%")

                # Optional: plot for head 0
                if h == 0:
                    plt.figure(figsize=(12, 5))
                    plt.subplot(1, 2, 1)
                    plt.title("Diff: Softmax vs TopK Mask")
                    plt.imshow(diff_topk, cmap='hot')
                    plt.colorbar()
                    plt.subplot(1, 2, 2)
                    plt.title("Diff: Softmax vs Memory Mask")
                    plt.imshow(diff_mem, cmap='hot')
                    plt.colorbar()
                    plt.tight_layout()
                    plt.show()

            print(f"\n📊 Average Accuracy (TopK Mask): {np.mean(Accuracy_list):.2f}%")
            print(f"📊 Average Accuracy (Memory Mask): {np.mean(Memory_Mask_Accuracy_list):.2f}%")

        else:
            print("Invalid model choice! Please choose a number from 1 to 7.")

    except ValueError as e:
        print("Error:", e)
